struct unit_s Nr[NRSTACKLEN],*nr,retzero;

struct unit_s unitlist[] = {
  {"m",1,{1}},
  {"kg",1,{0,1}}, /* bug: kkg allowed */
  {"s",1,{0,0,1}},
  {"K",1,{0,0,0,1}},
  {"mol",1,{0,0,0,0,1}},
  {"A",1,{0,0,0,0,0,1}},
  {"g",1e-3,{0,1}}, /* to parse mg etc. */
  {"ct",0.2e-3,{0,1}}, /* carat */
  //  {"cd",1,{0,0,0,0,0,0,1}},
  //  {"sr",1,{0,0,0,0,0,0,0,1}},

  {"J",1,{2,1,-2}},
  {"W",1,{2,1,-3}},
  {"Wh",3600,{2,1,-2}},
  {"N",1,{1,1,-2}},
  {"Pa",1,{-1,1,-2}},
  {"P",0.1,{-1,1,-1}}, // Poise = g.cm-1.s-1 
  {"C",1,{0,0,1,0,0,1}},
  {"V",1,{2,1,-3,0,0,-1}}, // volt = m2·kg·s−3·A−1
  {"S",1,{-2,-1,3,0,0,2}}, // siemens = kg−1·m−2·s3·A2
  {"Mho",1,{-2,-1,3,0,0,2}}, // siemens = kg−1·m−2·s3·A2
  {"Ohm",1,{2,1,-3,0,0,-2}}, // Ohm = kg·m2·s-3·A-2
  //  {"Ω",1,{2,1,-3,0,0,-2}}, // Ω = kg−1·m−2·s3·A2
  {"F",1,{-2,-1,4,0,0,2}}, // Farad = 1 s4·A2·m−2·kg−1
  {"H",1,{2,1,-2,0,0,-2}}, // 1 Henry = kg⋅m2⋅s-2⋅A-2
  {"T",1,{0,1,-2,0,0,-1}}, // 1 Tesla = kg·s−2·A−1
  {"Wb",1,{2,1,-2,0,0,-1}}, // 1 Weber = 1 kg·m2·s-2·A-1

  {"D",1e-21/299792458,{1,0,1,0,0,1}}, // Debye

  {"min",60,{0,0,1}},
  {"h",3600,{0,0,1}},
  {"d",3600*24,{0,0,1}},
  {"Ci",3.7e10,{0,0,-1}},
  {"Hz",1,{0,0,-1}},
  {"a_t",31556925.216,{0,0,1}}, /* annus = mean tropical year */
  {"a_j",31557600,{0,0,1}}, /* annus = Julian year */
  {"a_g",31556952,{0,0,1}}, /* annus = Gregorian year */
  {"a",  31556925.445,{0,0,1}}, /* annus = tropical year 2000 */
  {"Bq",1,{0,0,-1}},
  {"L",1e-3,{3}},
  {"gal",3.785411784e-3,{3}}, /* 231 in^3 */
  {"USgal",3.785411784e-3,{3}},
  {"UKgal",4.54609e-3,{3}},
  {"M",1e-3,{-3,0,0,0,1}}, /* mol/L */
  {"Da",1e-3,{0,1,0,0,-1}}, /* g/mol */
  {"AA",1e-10,{1,0,0,0,0,0}},
  {"eV",1.6021766208e-19,{2,1,-2}},
  {"cal",4.184,{2,1,-2}},
  {"calIT",4.1868,{2,1,-2}},
  {"calIUNS",4.182,{2,1,-2}},
  {"in",0.0254,{1}},
  {"ft",12*0.0254,{1}},
  {"yd",36*0.0254,{1}},
  {"mi",1760*36*0.0254,{1}},
  {"nmi",1852,{1}},
  {"BTU",1055.056,{2,1,-2}},
  {"HP",745.69987158227,{2,1,-3}}, /* mechanical, imperial */
  //  {"HPe",746,{2,1,-3}}, /* electric */
  {"PS",735.49875,{2,1,-3}}, /* metric */
  {"lb",0.45359237,{0,1}},
  {"PSI",6894.757293168,{-1,1,-2}},
  {"bar",1e5,{-1,1,-2}},
  {"atm",101325.,{-1,1,-2}},
  {"at",98066.5,{-1,1,-2}},
  {"torr",101325./760,{-1,1,-2}},
  {"foe",1e44,{2,1,-2}}, /* 10 to fifty-one erg */

  {"t",1e3,{0,1}}, /* metric ton - must be after "ft" */
  {"au",149597870700,{1}}, /* astronomical unit */
  {"erg",1e-7,{2,1,-2}},
  {"dyn",1e-5,{1,1,-2}},
  {NULL,0,{0}}};

struct pref_s {
  char *name;
  double prefix;
} prefixes[] = {
  {"",1},
  {"y",1e-24},
  {"z",1e-21},
  {"a",1e-18},
  {"f",1e-15},
  {"p",1e-12},
  {"n",1e-9},
  {"u",1e-6},
  //  {"μ",1e-6},
  {"m",1e-3},
  {"c",1e-2},
  {"d",1e-1},
  {"da",10},
  {"h",100},
  {"k",1e3},
  {"ki",1024.},
  {"M",1e6},
  {"Mi",1024.*1024.},
  {"G",1e9},
  {"Gi",1024.*1024.*1024.},
  {"T",1e12},
  {"Ti",1024.*1024.*1024.*1024.},
  {"P",1e15},
  {"E",1e18},
  {"Z",1e21},
  {"Y",1e24},
  {NULL,0}};

void parseunit(struct unit_s *nr,char *b,char **end)
/*
   parses "unit]", return value in nr->pow[]
   examples:
     [ms]=1e-3 s
     [m s]=m*s
     [m.s]=m^0*s (because "." expands to "0.0", which is zero)
     [m1/2]=[m0.5]=m^0.5
     [kg/m s]=[kg/m/s]=kg/(m*s)
     [us]=1e-6 s
*/
{
  int i,j,k,sg=1;
  char *a,*e;
  double prefix=1;


  loop (k,0,CALC_NUNITS) nr->pow[k]=0;

  /* in case of bad syntax */
  e=strchr(b,']');
  if (!e) e=b+strlen(b)-1;

  while (*b && *b!=']') {
    double ex=1; /* exponent of unit */

    while (*b && *b<=' ') b++;
    /* test divide by unit (not part of fractional exponent)
       [m/s/kg]=[m s-1 kg-1] */
    if (*b=='/') sg=-1,b++;
    /* this version would interpret [m/s/kg]=[m s-1 kg]:
       if (*b=='/') sg=-sg,b++; */
    while (*b && *b<=' ') b++;

    if (!isalpha(*b)) {
      fprintf(stderr,"calcu: syntax [%s\n",b);
      return; }

    for (a=b; isalpha(*a); a++);
    /* unit (w. prefix) = b to a-1 incl. */

    /* optional exponent expected at a */
    ex=strtod(a,&e);
    if (e==a)
      /* no exponent -> 1 */
      ex=1;
    else {
      /* exponent found */
      while (*e && *e<=' ') e++;
      if (*e=='/') {
        /* fractional exponent ? */
        char *ee;
        double den;

        den=strtod(e+1,&ee);
        if (e+1!=ee) {
          /* yes, fractional exponent */
          ex/=den;
          e=ee; } } }

    for (i=0; unitlist[i].name; i++) {
      char *unitbeg=a-strlen(unitlist[i].name); /* if found */

      if (!memcmp(unitbeg,unitlist[i].name,strlen(unitlist[i].name))) {

        j=0; /* no prefix -> 1 */
        if (unitbeg==b) goto doprefix;
        if (unitbeg>b)
          for (j=1; prefixes[j].name; j++)
            if (!memcmp(b,prefixes[j].name,unitbeg-b)) goto doprefix;
        /* prefix not found, trying name (e.g., atm: is not at+m) */
        continue;
      doprefix:
        loop (k,0,CALC_NUNITS) nr->pow[k]+=unitlist[i].pow[k]*ex*sg;
        prefix*=pow(unitlist[i].val*prefixes[j].prefix,ex*sg);
        goto found; } }
    fprintf(stderr,"calcu: unknown unit ignored [%s\n",b);
    /* not found - ignore for now */
  found:;
    /* NB: syntax with a dot [J.s] solved here: */
    while (*e && (*e<=' ' || *e=='.')) e++;
    b=e; }

  if (!*e) *end=e; /* missing ] */
  else *end = e+1;
  nr->val*=prefix;

  return;
}

char *unitserr;

void opunits(struct unit_s *a,struct unit_s *b,char assignopb)
/*
   combines units of two operands, checks if permitted
   a=1st operand and result, b=2nd operand
   E.g.: [m]+[kg] is invalid, [m]*[kg]=[m kg], [m]=[m] is [1],...
*/
{
  int k;
  unitserr=NULL;

  loop (k,0,CALC_NUNITS) switch (assignopb) {
    case 'p':
      /* parm[]: check */
      if (a->pow[k]!=b->pow[k]) unitserr="units of parameters do not match\n";
      break;
    case '+':
    case '-':
      /* sum, etc.: check if units match */
      if (a->pow[k]!=b->pow[k]) unitserr="cannot add,subtract different units\n";
      break;
    case '<':
    case '>':
      /* relation operators: check + set dimensionless */
      if (a->pow[k]!=b->pow[k]) unitserr="cannot compare different units\n";
      a->pow[k]=0;
      break;
    case 'f':
      /* check no unit */
      if (a->pow[k]!=0) unitserr="function argument is not dimensionless\n";
      break;
    case 'i':
      /* check no unit */
      if (a->pow[k]!=0) unitserr="BUG: iter,repeat,contfrac require dimensionless args\n";
      break;
    case '*': a->pow[k]+=b->pow[k]; break;
    case '/': a->pow[k]-=b->pow[k]; break;
    case '\\': a->pow[k]/=2; break; // square root
    case '^':
      /* power */
      if (b->pow[k]!=0) unitserr="power is not dimensionless\n";
      a->pow[k]*=b->val;
      break;
    case '0':
      /* set zero */
      a->pow[k]=0;
      break;
    default: fprintf(stderr,"internal: bad opunits key\n");
  }

  if (unitserr) fputs(unitserr,stderr);
}

/* user-friendly units - if found, this unit will be used */
struct user_s {
  struct user_s *next;
  char *unit;
} *userhead;

void removeunit(struct user_s *u) /****************************** removeunit */
{
  struct user_s *uu;

  if (!u) return;

  if (u==userhead) {
    userhead=userhead->next;
    free(u);
    return; }
  else
    looplist (uu,userhead) if (uu->next==u) {
      uu->next=u->next;
      free(u);
      return; }
}

void towards(char *u) /********************************************* towards */
/* command to, u points to the argument
  register preferred form of unit: to J/s, to [J/s]
  remove it: to ~J/s, to ~[J/s]
  remove all: to ~
 */
{
  struct user_s *user;
  int rm=0;
  char *ket;

  while (*u==' ' || *u=='\t') u++;

  if (*u==0) return;

  if (*u=='~') rm=1,u++;
  while (*u==' ' || *u=='\t') u++;

  if (*u==0) { /* remove all */
    struct user_s *uu,*next;

    for (uu=userhead; uu; uu=next) {
      next=uu->next;
      free(uu); }
    userhead=NULL;

    return; }

  if (*u=='[') u++;

  ket=strchr(u,']');
  if (ket) *ket=0;

  if (rm) {
    looplist (user,userhead) if (!strcmp(user->unit,u)) {
        removeunit(user);
        return; }
    /* no error message if not found */
    return; }

  user=malloc(sizeof(*user));
  if (!user) { fprintf(stderr,"no heap\n"); exit(1); }
  user->next=userhead;
  userhead=user;
  user->unit=strdup(u);
}

double unitfactor=1;

char *unit(struct unit_s *res) /*************************************** unit */
/* return unit as a string */
{
  static char prtunit[128];
  char *p=prtunit+1;
  int k;
  int nz=0;
  struct unit_s nr;
  struct user_s *u;

  looplist (u,userhead) {
    char *end;

    prtunit[0]='['; strcpy(p,u->unit); strcat(p,"]");
    nr.val=1;
    parseunit(&nr,p,&end);
    loop (k,0,CALC_NUNITS) if (res->pow[k]!=nr.pow[k]) goto notfound;
    unitfactor=nr.val;
    return prtunit;
  notfound: ; }

  unitfactor=1;
  *p=0;
  loop (k,0,CALC_NUNITS) if (res->pow[k]) {
    nz++;
    if (res->pow[k]==1)
      p+=sprintf(p,"%s ",unitlist[k].name);
    else
      p+=sprintf(p,"%s%g ",unitlist[k].name,res->pow[k]); }
  if (nz) {
    prtunit[0]='[';
    p[-1]=']'; }
  else
    prtunit[0]=0;

  return prtunit;
}
